/*
**    TP 4ETI CPE Lyon
**    Copyright (C) 2012 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



#include <cstdlib>
#include <fstream>
#include <cmath>


#include <string_helper.hpp>
#include <mesh.hpp>
#include <matrix3.hpp>

#include <mesh_io_off.hpp>


namespace cpe
{
    mesh::mesh(){}
    mesh::mesh(const std::pair <std::vector<v3>,std::vector<unsigned int> >& input)
    {
        vertex=input.first;
        connectivity=input.second;
    }

    v3 mesh::get_vertex(const unsigned int& k_tri,const unsigned int& k_v) const
    {
        assert_connectivity(k_tri);
        if(k_v>2)
        {throw std::string("mesh::get_vertex("+cpe::string_converter::to_string(k_tri)+","+cpe::string_converter::to_string(k_v)+"), k_v must be < 2");}
        return get_vertex(connectivity[3*k_tri+k_v]);
    }
    v3 mesh::get_vertex(const unsigned int& k_vertex) const
    {
        assert_vertex(k_vertex);
        return vertex[k_vertex];
    }
    v3& mesh::get_vertex(const unsigned int& k_tri,const unsigned int& k_v)
    {
        assert_connectivity(k_tri);
        if(k_v>2)
        {throw std::string("mesh::get_vertex("+cpe::string_converter::to_string(k_tri)+","+cpe::string_converter::to_string(k_v)+"), k_v must be < 2");}
        return get_vertex(connectivity[3*k_tri+k_v]);
    }
    v3& mesh::get_vertex(const unsigned int& k_vertex)
    {
        assert_vertex(k_vertex);
        return vertex[k_vertex];
    }
    std::vector<unsigned int> mesh::get_connectivity(const unsigned int& k_polygon) const
    {
        assert_connectivity(k_polygon);
        std::vector<unsigned int> index(3);
        for(unsigned int k=0;k<3;++k)
            index[k]=connectivity[3*k_polygon+k];
        return index;
    }

    void mesh::set_vertex(const unsigned int& k_vertex,const v3& new_vertex)
    {
        assert_vertex(k_vertex);
        vertex[k_vertex]=new_vertex;
    }
    void mesh::set_connectivity(const unsigned int& k_polygon,const std::vector<unsigned int>& new_triangle)
    {
        assert_connectivity(k_polygon);
        if(new_triangle.size()<3)
        {std::cout<<"Error in mesh::set_connectivity, size of triangle is "<<new_triangle.size()<<std::endl;exit(-1);}

        for(unsigned int k_dim=0;k_dim<3;++k_dim)
            connectivity[3*k_polygon+k_dim]=new_triangle[k_dim];

    }


    const std::vector<v3>& mesh::get_vertex() const
    {return vertex;}
    std::vector<v3>& mesh::get_vertex()
    {return vertex;}

    const std::vector <unsigned int>& mesh::get_connectivity() const
    {return connectivity;}
    std::vector <unsigned int>& mesh::get_connectivity()
    {return connectivity;}


    mesh mesh::load_off_file(const std::string& filename)
    {
        return mesh_io_off::load_mesh_io_off(filename);
    }
    void mesh::save_off_file(const std::string& filename) const
    {
        mesh_io_off::save_mesh_io_off(filename,*this);
    }
    unsigned int mesh::vertex_number() const{return vertex.size();}
    unsigned int mesh::triangle_number() const{return connectivity.size()/3;}

    void mesh::scale_unity()
    {
        if(vertex_number()==0)
            return;

        v3 x_max=vertex[0];
        v3 x_min=x_max;

        //get bounding box
        for(unsigned int k=0,N=vertex_number();k<N;++k)
        {
            v3 x=vertex[k];
            for(unsigned int k_dim=0;k_dim<3;++k_dim)
            {
                if(x[k_dim]<x_min[k_dim])
                    x_min[k_dim]=x[k_dim];
                if(x[k_dim]>x_max[k_dim])
                    x_max[k_dim]=x[k_dim];
            }
        }

        //scale
        v3 x_size=x_max-x_min;
        double x_size_max=std::max(std::max(x_size[0],x_size[1]),x_size[2]);

        double epsilon=1e-5;
        if(std::fabs(x_size_max)<epsilon)
        {
            std::cout<<"Scaling failed, mesh is too small in mesh::scale_unity"<<std::endl;
            return;
        }

        v3 x_middle=(x_max+x_min)/2.0;
        for(unsigned int k=0,N=vertex_number();k<N;++k)
            vertex[k]=(vertex[k]-x_middle)/x_size_max;


    }

    std::map<int,std::list<int> > mesh::ring() const
    {
        std::map<int,std::list<int> > r;
        for(int k=0,N=triangle_number();k<N;++k)
        {
            for(int k_dim=0;k_dim<3;++k_dim)
            {
                int u0=connectivity[3*k+k_dim];
                int u1=connectivity[3*k+(k_dim+1)%3];

                r[u0].push_back(u1);
                r[u1].push_back(u0);
            }
        }

        return r;
    }
    std::map<int,std::list<int> > mesh::star() const
    {
        std::map<int,std::list<int> > s;
        for(unsigned int k=0,N=triangle_number();k<N;++k)
            for(unsigned int k_dim=0;k_dim<3;++k_dim)
                s[connectivity[3*k+k_dim]].push_back(k);

        return s;
    }

    std::vector <v3> mesh::normal_vertex() const
    {
        std::map<int,std::list<int> > s=star();
        std::vector <v3> n(vertex.size());
        std::vector <v3> n_poly=normal_polygon();

        for(unsigned int k=0,N=vertex_number();k<N;++k)
        {
            v3 temp_normal;
            for(std::list<int>::const_iterator it=s[k].begin(),it_end=s[k].end();it!=it_end;++it)
                temp_normal += n_poly[*it];

            temp_normal=temp_normal.normalized();
            n[k]=temp_normal;
        }

        return n;

    }
    std::vector <v3> mesh::normal_polygon() const
    {
        std::vector <v3> n(connectivity.size()/3);
        for(unsigned int k=0,N=triangle_number();k<N;++k)
        {
            v3 x0=v3(vertex[connectivity[3*k+0]]);
            v3 x1=v3(vertex[connectivity[3*k+1]]);
            v3 x2=v3(vertex[connectivity[3*k+2]]);

            v3 xn=((x1-x0).normalized().cross((x2-x0).normalized())).normalized();

            n[k]=xn;
        }
        return n;
    }

    void mesh::add_vertex(const v3& x)
    {
        vertex.push_back(x);
    }
    void mesh::add_vertex(const double& x,const double& y,const double& z)
    {
        vertex.push_back(v3(x,y,z));
    }

    void mesh::add_triangle(const unsigned int& u0,const unsigned int& u1,const unsigned int& u2)
    {
        connectivity.push_back(u0);connectivity.push_back(u1);connectivity.push_back(u2);
    }

    void mesh::assert_vertex(const unsigned int& k_vertex) const
    {if(k_vertex>=vertex.size()){std::cout<<"Error in mesh::assert_vertex("<<k_vertex<<"), vertex.size()="<<vertex.size()<<std::endl;throw std::string("mesh::assert_vertex");}}
    void mesh::assert_connectivity(const unsigned int& k_connectivity) const
    {if(3*k_connectivity+2>=connectivity.size()){std::cout<<"Error in mesh::assert_connectivity("<<k_connectivity<<"), connectivity.size()="<<connectivity.size()<<std::endl;throw std::string("mesh::assert_connectivity");}}

    void mesh::inverse_triangle_order()
    {
        for(unsigned int k=0,N=triangle_number();k<N;++k)
        {
            unsigned int temp=connectivity[3*k+1];
            connectivity[3*k+1]=connectivity[3*k+0];
            connectivity[3*k+0]=temp;
        }
    }

    const std::vector<double>& mesh::get_texture() const
    {
        return texture;
    }

    std::vector<double>& mesh::get_texture()
    {
        return texture;
    }

    void mesh::set_texture(const unsigned int& k_vertex,const double& tx,const double& ty)
    {
        if(2*k_vertex>=texture.size())
            throw std::string("mesh::set_texture("+string_converter::to_string(k_vertex)+"), size="+string_converter::to_string(texture.size()/2));
        texture[2*k_vertex+0]=tx;
        texture[2*k_vertex+1]=ty;

    }

    double mesh::get_texture_tx(const unsigned int& k_vertex) const
    {
        if(2*k_vertex>=texture.size())
            throw std::string("mesh::get_texture_tx("+string_converter::to_string(k_vertex)+"), size="+string_converter::to_string(texture.size()/2));
        return texture[2*k_vertex+0];
    }
    double mesh::get_texture_ty(const unsigned int& k_vertex) const
    {
        if(2*k_vertex>=texture.size())
            throw std::string("mesh::get_texture_ty("+string_converter::to_string(k_vertex)+"), size="+string_converter::to_string(texture.size()/2));
        return texture[2*k_vertex+1];
    }

    void mesh::scale(const double& value)
    {
        for(unsigned int k=0,N=vertex.size();k<N;++k)
            vertex[k]*=value;
    }
    void mesh::apply(const matrix3& M)
    {
        for(unsigned int k=0,N=vertex_number();k<N;++k)
            vertex[k]=M*vertex[k];
    }

    void mesh::translate(const v3& tr)
    {
        for(unsigned int k=0,N=vertex_number();k<N;++k)
            vertex[k] += tr;
    }

}
