/*
**    TP 4ETI CPE Lyon
**    Copyright (C) 2012 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



#include <GL/glut.h>


#include <opengl_drawer.hpp>
#include <texture.hpp>
#include <grid.hpp>
#include <surface_param_drawable.hpp>

namespace cpe
{


    void opengl_drawer::draw(const mesh& m)
    {
        draw_edges(m);
    }

    void opengl_drawer::draw(const mesh& m,const std::vector<v3>& normal)
    {
        if(normal.size()==m.triangle_number())
            draw_per_polygon_normal(m,normal);
        else if(normal.size()==m.vertex_number())
            draw_per_vertex_normal(m,normal);
    }

    void opengl_drawer::draw_per_polygon_normal(const mesh& m,const std::vector<v3>& normal)
    {
        const unsigned int* p_connectivity=&m.get_connectivity()[0];
        const double* p_vertex=&(m.get_vertex()[0].x());
        const double* p_normal=&(normal[0].x());
        const double* p_texture=&m.get_texture()[0];
        bool is_texture=(m.get_texture().size() == 2*m.vertex_number());




        glBegin(GL_TRIANGLES);

        unsigned int N_tri=m.triangle_number();

        if(is_texture==false)
        {
            for(unsigned int k_tri=0;k_tri<N_tri;++k_tri)
            {
                //set normal
                glNormal3dv(p_normal+3*k_tri);

                //draw vertices
                for(unsigned int k_dim=0;k_dim<3;++k_dim)
                    glVertex3dv(p_vertex+3*(*(p_connectivity++)));
            }
        }
        else
        {
            for(unsigned int k_tri=0;k_tri<N_tri;++k_tri)
            {
                //set normal
                glNormal3dv(p_normal+3*k_tri);

                //draw vertices
                for(unsigned int k_dim=0;k_dim<3;++k_dim)
                {
                    glTexCoord2dv(p_texture+2*(*(p_connectivity  )));
                    glVertex3dv  (p_vertex +3*(*(p_connectivity++)));
                }
            }
        }

        glEnd();
    }

    void opengl_drawer::draw_per_vertex_normal(const mesh& m,const std::vector<v3>& normal)
    {
        unsigned int N_tri=m.triangle_number();
        const unsigned int* p_connectivity=&m.get_connectivity()[0];
        const double* p_vertex=&(m.get_vertex()[0].x());
        const double* p_normal=&(normal[0].x());
        const double* p_texture=&m.get_texture()[0];
        bool is_texture=(m.get_texture().size() == 2*m.vertex_number());


        //enable client state
        glEnableClientState(GL_VERTEX_ARRAY);
        glEnableClientState(GL_NORMAL_ARRAY);
        if(is_texture)
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);

        //setup pointers
        glVertexPointer(3,GL_DOUBLE,0,p_vertex);
        glNormalPointer(GL_DOUBLE,0,p_normal);
        if(is_texture)
            glTexCoordPointer(2,GL_DOUBLE,0,p_texture);

        //draw triangles
        glDrawElements(GL_TRIANGLES,3*N_tri,GL_UNSIGNED_INT,p_connectivity);

        //disable client state
        glDisableClientState(GL_VERTEX_ARRAY);
        glDisableClientState(GL_NORMAL_ARRAY);
        if(is_texture)
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);

    }

    void opengl_drawer::draw_edges(const mesh& m)
    {
        glDisable(GL_LIGHTING);

        const unsigned int* p_connectivity=&m.get_connectivity()[0];
        const double* p_vertex=&(m.get_vertex()[0].x());


        unsigned int N_tri=m.triangle_number();

        //draw edge
        glBegin(GL_LINES);
        for(unsigned int k_tri=0;k_tri<N_tri;++k_tri)
        {
            for(unsigned int k_dim=0;k_dim<3;++k_dim)
            {
                glVertex3dv(p_vertex+3*p_connectivity[3*k_tri+k_dim]);
                glVertex3dv(p_vertex+3*p_connectivity[3*k_tri+(k_dim+1)%3]);
            }
        }
        glEnd();


    }



    void opengl_drawer::draw_normals(const mesh& m,const std::vector<v3>& normal,const double& L_scaling)
    {
        glDisable(GL_LIGHTING);



        unsigned int N_tri=m.triangle_number();

        //draw per face normal
        if(normal.size()==m.triangle_number())
        {
            glBegin(GL_LINES);
            for(unsigned int k_tri=0;k_tri<N_tri;++k_tri)
            {
                //get barycenter of the triangle
                v3 v0=m.get_vertex(k_tri,0); v3 v1=m.get_vertex(k_tri,1); v3 v2=m.get_vertex(k_tri,2);
                v3 bar0 = 1.0/3.0*(v0+v1+v2);

                //get current normal
                const v3& n=normal[k_tri];

                //extremity
                v3 bar1=bar0+0.1*L_scaling*n;


                // draw normals
                glVertex3d(bar0.x(),bar0.y(),bar0.z());
                glVertex3d(bar1.x(),bar1.y(),bar1.z());

            }
            glEnd();
        }
        //draw per vertex normal
        else if(normal.size()==m.vertex_number())
        {
            const double* p_vertex=&(m.get_vertex()[0].x());
            const double* p_normal=&(normal[0].x());

            glBegin(GL_LINES);
            for(unsigned int k_v=0,N_v=m.vertex_number();k_v<N_v;++k_v)
            {


                //starting point
                glVertex3dv(p_vertex);
                //ending point
                glVertex3d(p_vertex[0]+0.1*L_scaling*p_normal[0],
                           p_vertex[1]+0.1*L_scaling*p_normal[1],
                           p_vertex[2]+0.1*L_scaling*p_normal[2]);

                p_vertex+=3;
                p_normal+=3;
            }
            glEnd();
        }



    }

    void opengl_drawer::generate_texture(const unsigned int& id_texture,const std::string& filename)
    {
        //load texture data
        texture t;
        t.load_picture(filename);

        GLuint id=id_texture;

        //generate texture id
        if(glIsTexture(id_texture))
            glDeleteTextures(1,&id);
        glGenTextures(1,&id);
        glBindTexture(GL_TEXTURE_2D,id);

        //build mipmap and store texture
        gluBuild2DMipmaps( GL_TEXTURE_2D, 4, t.get_Nx(),t.get_Ny(), GL_RGBA, GL_UNSIGNED_BYTE, &t.get_data()[0] );
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    }

    void opengl_drawer::draw(const grid& g)
    {
        const unsigned int& Nu=g.size_u();
        const unsigned int& Nv=g.size_v();

        glBegin(GL_LINES);
        for(unsigned int ku=0;ku<Nu-1;++ku)
        {
            for(unsigned int kv=0;kv<Nv;++kv)
            {
                const v3& x0=g(ku,kv);
                const v3& x1=g(ku+1,kv);
                glVertex3dv(&x0.x());
                glVertex3dv(&x1.x());
            }
        }
        for(unsigned int ku=0;ku<Nu;++ku)
        {
            for(unsigned int kv=0;kv<Nv-1;++kv)
            {
                const v3& x0=g(ku,kv);
                const v3& x1=g(ku,kv+1);
                glVertex3dv(&x0.x());
                glVertex3dv(&x1.x());
            }
        }
        glEnd();


    }

    void opengl_drawer::draw(const surface_param_drawable& surf)
    {
        const unsigned int& Nu=surf.size_u();
        const unsigned int& Nv=surf.size_v();
        unsigned int N_quad=(Nu-1)*(Nv-1);

        const double* p=surf.pointer_draw_vertices();
        const double* n=surf.pointer_draw_normal();
        const unsigned int* c=surf.pointer_draw_connectivity();


        //enable client state
        glEnableClientState(GL_VERTEX_ARRAY);
        glEnableClientState(GL_NORMAL_ARRAY);

        //setup pointers
        glVertexPointer(3,GL_DOUBLE,0,p);
        glNormalPointer(GL_DOUBLE,0,n);

        //draw triangles
        glDrawElements(GL_QUADS,4*N_quad,GL_UNSIGNED_INT,c);

        //disable client state
        glDisableClientState(GL_VERTEX_ARRAY);
        glDisableClientState(GL_NORMAL_ARRAY);

    }

    void opengl_drawer::draw_curve(const std::vector<std::vector<cpe::v3> >& curves)
    {
        glBegin(GL_LINES);
        const unsigned int& N_curve=curves.size();
        for(unsigned int k=0;k<N_curve;++k)
        {
            const std::vector<cpe::v3>& current_curve=curves[k];
            const unsigned int& N_sample=current_curve.size();
            for(unsigned int k2=0;k2<N_sample-1;++k2)
            {
                v3 x0=current_curve[k2];
                v3 x1=current_curve[k2+1];

                glVertex3d(x0.x(),x0.y(),x0.z());
                glVertex3d(x1.x(),x1.y(),x1.z());
            }
        }
        glEnd();
    }

}



