/*
**    TP 4ETI CPE Lyon
**    Copyright (C) 2012 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef _GRID_HPP_
#define _GRID_HPP_

#include <v3.hpp>
#include <vector>

namespace cpe
{

class patch4;

class grid
{
    public:

        // ********************************************* //
        // ********************************************* //
        //  CONSTRUCTORS
        // ********************************************* //
        // ********************************************* //

        /** \brief empty constructor */
        grid();
        /** \brief direct constructor with size */
        grid(const unsigned int& Nu,const unsigned int& Nv);

        // ********************************************* //
        // ********************************************* //
        //  Size
        // ********************************************* //
        // ********************************************* //

        /** \brief resize the grid */
        void resize(const unsigned int& Nu,const unsigned int& Nv);

        /** \brief return the size in u direction */
        const unsigned int& size_u() const;
        /** \brief return the size in v direction */
        const unsigned int& size_v() const;


        // ********************************************* //
        // ********************************************* //
        //  Data
        // ********************************************* //
        // ********************************************* //

        /** \brief Accessor to the value (ku,kv) */
        const v3& operator()(const unsigned int& ku,const unsigned int& kv) const;
        /** \brief Accessor to the value (ku,kv) */
        v3& operator()(const unsigned int& ku,const unsigned int& kv);


        // ********************************************* //
        // ********************************************* //
        //  4x4 patch
        // ********************************************* //
        // ********************************************* //

        /** \brief get a 4x4 patch */
        patch4 get_patch(const unsigned int& k_patch_u,const unsigned int& k_patch_v) const;


        // ********************************************* //
        // ********************************************* //
        //  Special initialization
        // ********************************************* //
        // ********************************************* //

        /** \brief build Matlab meshgrid style from 1 to size */
        void build_meshgrid();
        /** \brief build spherical distribution */
        void build_sphere();


        // ********************************************* //
        // ********************************************* //
        //  Grid transformation
        // ********************************************* //
        // ********************************************* //

        /** \brief scale the grid homogeneously by a scalar */
        grid& operator*=(const double& s);
        /** \brief scale the grid homogeneously by a scalar */
        friend grid operator*(const double& s,const grid& g);
        /** \brief scale the grid homogeneously by a scalar */
        friend grid operator*(const grid& g,const double& s);

        /** \brief translate the grid */
        grid& operator+=(const v3& t);
        /** \brief translate the grid */
        grid& operator-=(const v3& t);
        /** \brief translate the grid */
        friend grid operator+(const grid& g,const v3& t);
        /** \brief translate the grid */
        friend grid operator+(const v3& t,const grid& g);
        /** \brief translate the grid */
        friend grid operator-(const grid& g,const v3& t);
        /** \brief translate the grid */
        friend grid operator-(const v3& t,const grid& g);

        /** \brief negation positions of the grid */
        grid operator-() const;


        // ********************************************* //
        // ********************************************* //
        //  Topologie transform
        // ********************************************* //
        // ********************************************* //

        /** \brief add a line of grid in the u direction at the end */
        void add_back_u();
        /** \brief add a line of grid in the u direction at the beginning */
        void add_front_u();

        /** \brief add a line of grid in the v direction at the end */
        void add_back_v();
        /** \brief add a line of grid in the v direction at the beginning */
        void add_front_v();

        /** \brief suppress a line of grid in the u direction at the end */
        void suppress_back_u();
        /** \brief suppress a line of grid in the u direction at the beginning */
        void suppress_front_u();

        /** \brief suppress a line of grid in the v direction at the end */
        void suppress_back_v();
        /** \brief suppress a line of grid in the v direction at the beginning */
        void suppress_front_v();

        /** \brief duplicate positions at the boundaries */
        void duplicate_boundary();


        // ********************************************* //
        // ********************************************* //
        //  INPUT OUTPUT
        // ********************************************* //
        // ********************************************* //

        /** \brief Write grid into a file */
        void write(const std::string& filename);
        /** \brief Read grid from a file */
        static grid read(const std::string& filename);

    private:

        /** \brief ensure (ku,kv) is coherent with the size of the grid */
        void assert_size(const unsigned int& ku,const unsigned int& kv) const;

        /** \brief internal size u */
        unsigned int internal_size_u;
        /** \brief internal size v*/
        unsigned int internal_size_v;

        /** \brief Internal storage under the form: ku+Nu*kv */
        std::vector<v3> internal_data;
};
}

#endif
