/*
**    TP 4ETI CPE Lyon
**    Copyright (C) 2012 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


/** SCENE HEADER */


#ifndef _SCENE_HPP_
#define _SCENE_HPP_

#include <v3.hpp>
#include <mesh.hpp>
#include <grid.hpp>
#include <surface_param_drawable.hpp>
#include <picking_data.hpp>
#include <gui_interface.hpp>

#include <vector>

class glwidget;

class scene
{
    private:

        /** \brief The control grid for the spline */
        cpe::grid g0;
        /** \brief The Spline surface */
        cpe::surface_param_drawable surf;

    public:

        // ********************************************* //
        // ********************************************* //
        //  CONSTRUCTORS
        // ********************************************* //
        // ********************************************* //

        /** \brief Empty constructor for the scene */
        scene();


        // ********************************************* //
        // ********************************************* //
        //  Scene construction and drawing
        // ********************************************* //
        // ********************************************* //

        /** \brief Method called only once at the beginning (load files ...) */
        void load_model();
        /** \brief Method called at every frame */
        void draw_scene();

        // ********************************************* //
        // ********************************************* //
        //  Spline manager
        // ********************************************* //
        // ********************************************* //

        /** \brief Build a full spline surface of with a given sample rate */
        void build_spline_surface(const unsigned int& N_samples);
        /** \brief Fill a 4x4 patch by the surface */
        void fill_patch(const unsigned int& index_surface,const cpe::patch4& p,const unsigned int& N_samples);
        /** \brief Compute the curves at the boundary of each patch */
        void compute_curve_spline();

        /** \brief Update the surface given with the high sampling ratio */
        void update_high_res();
        /** \brief Update the surface given with the low sampling ratio (for interactive edition) */
        void update_low_res();



        // ********************************************* //
        // ********************************************* //
        //  Grid manager
        // ********************************************* //
        // ********************************************* //

        /** \brief Helper function for picking position of the grid */
        void picking_grid(const cpe::v3& center_ray,const cpe::v3& direction_ray);
        /** \brief Translate the selected position of the grid */
        void translate_picking(const cpe::v3 t);

        /** \brief add a u line at the end of the grid */
        void add_u_back_line();
        /** \brief add a v line at the end of the grid */
        void add_v_back_line();
        /** \brief add a u line at the front of the grid */
        void add_u_front_line();
        /** \brief add a v line at the front of the grid */
        void add_v_front_line();

        /** \brief suppress a u line at the end of the grid */
        void erase_u_back_line();
        /** \brief suppress a v line at the end of the grid */
        void erase_v_back_line();
        /** \brief suppress a u line at the beginning of the grid */
        void erase_u_front_line();
        /** \brief suppress a v line at the beginning of the grid */
        void erase_v_front_line();

        /** \brief duplicate each position of the grid at the boundary */
        void duplicate_boundary();



        // ********************************************* //
        // ********************************************* //
        //  Gui Helpers
        // ********************************************* //
        // ********************************************* //


        /** \brief Set the GLWidget pointer */
        void set_glwidget(glwidget *p_widget);

        /** \brief get the gui helper class */
        cpe::gui_interface& gui_visible_interface();



    private:

        /** \brief the color_table of the patches */
        std::vector<cpe::v3> color_table;

        /** \brief The curve of the boundaries */
        std::vector<std::vector<cpe::v3> > curve_spline;

        /** \brief storage for the sphere (to draw the grid position) */
        cpe::mesh sphere;
        /** \brief storage for the sphere normal (to draw the grid position) */
        std::vector<cpe::v3> sphere_normal;
        /** \brief Init the sphere loading */
        void preload_sphere();

        /** \brief storage for picking parameters */
        cpe::picking_data pick;

        /** \brief The sampling for the high resolution surface */
        unsigned int N_sample_high_res;
        /** \brief The sampling for the low resolution surface */
        unsigned int N_sample_low_res;


        /** \brief Storage class for visibility modified in the gui */
        cpe::gui_interface gui_visible;

        /** \brief counter to update the surface at some time */
        unsigned int counter_update;


        /** \brief Init the visibility gui parameters */
        void init_gui_visibility();
        /** \brief Init the colortable for the patches */
        void init_color_table();

};

#endif
