/*
**    TP 4ETI CPE Lyon
**    Copyright (C) 2012 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


/** SCENE.cpp */

#include <scene.hpp>

#include <GL/glew.h>
#include <QString>
#include <QFont>
#include <glwidget.hpp>


#include <opengl_drawer.hpp>
#include <texture.hpp>
#include <mesh_io_off.hpp>
#include <string_helper.hpp>
#include <cmath>
#include <matrix4.hpp>
#include <patch4.hpp>
#include <surface_param.hpp>
#include <surface_param_drawable.hpp>
#include <intersection_helper.hpp>
#include <evaluator_spline.hpp>
#include <matrix3.hpp>

using namespace cpe;





//************************************************************//
//*** Ici, on charge les objets de la scène une seule fois ***//
//************************************************************//
void scene::load_model()
{

    // *********************************************** //
    // Chargement d'une sphere sous forme de maillage
    // *********************************************** //
    preload_sphere();



    // *********************************************** //
    // Chargement de la grille de controle
    // *********************************************** //

    //definition des données suivant x
    matrix4 patch_x=matrix4(0,0,0,0,
                            1,1,1,1,
                            2,2,2,2,
                            3,3,3,3);

    //definition des données suivant y
    matrix4 patch_y=matrix4(0,1,2,3,
                            0,1,2,3,
                            0,1,2,3,
                            0,1,2,3);

    //definition des données suivant z
    matrix4 patch_z=matrix4(0,1,1,1.5,
                            1,2,2,1,
                            1,2,2,1,
                            1.5,1,1,-0.5);




    // *********************************************** //
    // Calcul du patch courant
    // *********************************************** //

    // Le foncteur d'evaluation en (x,y,z)
    evaluator_spline evaluateur_x(patch_x);
    evaluator_spline evaluateur_y(patch_y);
    evaluator_spline evaluateur_z(patch_z);

    //nombre d'échantillons suivant u et v
    unsigned int Nu=50;
    unsigned int Nv=50;

    //redimensionnement pour stocker la surface
    surf.resize(Nu,Nv);

    //Pour tous les echantillons suivant u
    for(unsigned int ku=0;ku<Nu;++ku)
    {
        double u=(double)(ku)/(Nu-1);

        //Pour tous les echantillons suivant v
        for(unsigned int kv=0;kv<Nv;++kv)
        {
            double v=(double)(kv)/(Nv-1);

            // ******************************** //
            // Evaluation des coordonnees       //
            // ******************************** //

            //Evaluation de la spline en (x,y,z)
            double x=evaluateur_x(u,v);
            double y=evaluateur_y(u,v);
            double z=evaluateur_z(u,v);

            //Coordonnee de mon point en ces parametres (u,v)
            surf(ku,kv)=v3(x,y,z);

            // ******************************** //
            // Evaluation de la normale         //
            // ******************************** //

            //Calcul de la derivee en u, puis en v
            v3 du=v3(evaluateur_x.diff_u(u,v),evaluateur_y.diff_u(u,v),evaluateur_z.diff_u(u,v));
            v3 dv=v3(evaluateur_x.diff_v(u,v),evaluateur_y.diff_v(u,v),evaluateur_z.diff_v(u,v));

            //Calcul de la normale
            v3 n=(du.cross(dv)).normalized();

            //Valeur de la normale en ce point
            surf.set_normal(ku,kv,n);
        }
    }

    //connectivitee de la surface
    surf.update_connectivity();

    //mise en forme de la grille de controle pour l'affichage
    g0.resize(4,4);
    for(unsigned int ku=0;ku<4;++ku)
        for(unsigned int kv=0;kv<4;++kv)
            g0(ku,kv)=v3(patch_x(ku,kv),patch_y(ku,kv),patch_z(ku,kv));

}



//************************************************************************//
//*** Ici, on affiche ce que l'on veut.                                ***//
//*** Cette fonction est appellee un grand nombre de fois par secondes ***//
//************************************************************************//
void scene::draw_scene()
{


    glEnable(GL_LIGHTING);//commande OpenGL pour l'eclairage


    try
    {

        //Affichage de la surface spline
        glEnable(GL_POLYGON_OFFSET_FILL);
        glPolygonOffset(6,6);

        glColor3d(0.5,0.9,0.8);

        //affiche de la surface proprement dit
        if(gui_visible.get_surface_visible()==true)
            opengl_drawer::draw(surf);

        glDisable(GL_POLYGON_OFFSET_FILL);


        //affichage de la grille de controle
        glLineWidth(1.5);
        glDisable(GL_LIGHTING);
        glColor3d(0.5,0.5,0.8);
        if(gui_visible.get_grid_visible()==true)
            opengl_drawer::draw(g0);
        glEnable(GL_LIGHTING);



        //affichage des positions de la grille en tant que sphere
        glEnable(GL_LIGHTING);
        if(gui_visible.get_gridposition_visible()==true)
        {
            const unsigned int& Nu=g0.size_u();
            const unsigned int& Nv=g0.size_v();
            for(unsigned int ku=0;ku<Nu;++ku)
            {
                for(unsigned int kv=0;kv<Nv;++kv)
                {
                    const v3& t=g0(ku,kv);

                    //translation rapide en OpenGL par Push/Pop-Matrix
                    glPushMatrix();
                    glTranslated(t.x(),t.y(),t.z());
                    glColor3d(1,0,0);

                    //modification de la couleur lors du picking d'une sphere
                    if(pick.get_is_picked()==true)
                    {
                        const std::map<unsigned int,std::pair<unsigned int,unsigned int> >& picked_index=pick.get_picked_index();
                        if(picked_index.find(ku+Nu*kv)!=picked_index.end())
                            glColor3d(1,1,0);
                    }

                    //affichage de la sphere
                    opengl_drawer::draw(sphere,sphere_normal);
                    glPopMatrix();
                }
            }
        }
        glDisable(GL_LIGHTING);


    }
    catch(std::string mess) //recuperation des exceptions
    {
        std::cout<<mess<<std::endl;
    }




}








//initialisation de la classe scene
scene::scene()
{
    init_gui_visibility();
}

//retourne l'interface de sauvegarde des parametres de la gui
gui_interface& scene::gui_visible_interface()
{
    return gui_visible;
}

//initialisation des parametres de visibilite de la scene
void scene::init_gui_visibility()
{
    gui_visible.set_grid_visible(true);
    gui_visible.set_gridposition_visible(true);
    gui_visible.set_surface_visible(true);
}
//prechargement de la sphere pour les sommets de la grille
void scene::preload_sphere()
{
    sphere=cpe::mesh::load_off_file("data/sphere.off");
    sphere.scale(0.06);
    sphere_normal=sphere.normal_vertex();
}

