/*
**    TP 4ETI CPE Lyon
**    Copyright (C) 2012 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _MATRIX4_HPP_
#define _MATRIX4_HPP_

#include <iostream>
#include <vector>

namespace cpe
{
class v4;
class v3;

class matrix4
{

    public:

        // ********************************************* //
        // ********************************************* //
        //  CONSTRUCTORS
        // ********************************************* //
        // ********************************************* //

        /** \brief empty constructor (identity) */
        matrix4();
        /** \brief direct constructor with every value */
        matrix4(const double& x00,const double& x01,const double& x02,const double& x03,
                const double& x10,const double& x11,const double& x12,const double& x13,
                const double& x20,const double& x21,const double& x22,const double& x23,
                const double& x30,const double& x31,const double& x32,const double& x33);

        // ********************************************* //
        // ********************************************* //
        //  Special initialization
        // ********************************************* //
        // ********************************************* //

        /** \brief build identity matrix */
        static matrix4 identity();
        /** \brief build zero matrix */
        static matrix4 zeros();
        /** \brief build rotation matrix */
        static matrix4 rotation(const v3& axis,const double& angle);
        /** \brief build scaling matrix (3D scaling, not w) */
        static matrix4 scale(const double& s);
        /** \brief build scaling matrix */
        static matrix4 scale(const double& s_x,const double& s_y,const double& s_z,const double& s_w);


        // ********************************************* //
        // ********************************************* //
        //  Element access
        // ********************************************* //
        // ********************************************* //

        /** \brief Access to the k_th entry (k in [0,3]) */
        const double& operator()(const size_t& k1,const size_t& k2) const;
        /** \brief Access to the k_th entry (k in [0,3]) */
        double& operator()(const size_t& k1,const size_t& k2);

        /** \brief fast pointer access
        \warning unprotected access */
        const double* pointer() const;
        /** \brief fast pointer access
        \warning unprotected access */
        double* pointer_unprotected();



        // ********************************************* //
        // ********************************************* //
        //  Math operator
        // ********************************************* //
        // ********************************************* //

        /** \brief + operator */
        friend matrix4 operator+(const matrix4& m1,const matrix4& m2);
        /** \brief + operator */
        friend matrix4 operator+(const double& s,const matrix4& m);
        /** \brief + operator */
        friend matrix4 operator+(const matrix4& m,const double& s);

        /** \brief - operator */
        friend matrix4 operator-(const matrix4& m1,const matrix4& m2);
        /** \brief - operator */
        friend matrix4 operator-(const double& s,const matrix4& m);
        /** \brief - operator */
        friend matrix4 operator-(const matrix4& m,const double& s);

        /** \brief multiply by a scalar operator */
        friend matrix4 operator*(const double& s,const matrix4& m);
        /** \brief multiply by a scalar operator */
        friend matrix4 operator*(const matrix4& m,const double& s);
        /** \brief multiply matrix with a v4 */
        friend v4 operator*(const matrix4& m,const v4& v);
        /** \brief multiply matrix with a v4 */
        friend v4 operator*(const v4& v,const matrix4& m);
        /** \brief matrix mulitplication */
        friend matrix4 operator*(const matrix4& m1,const matrix4& m2);
        /** \brief divide by a scalar operator */
        friend matrix4 operator/(const matrix4& m,const double& s);


        /** \brief internal + */
        matrix4& operator+=(const matrix4& m);
        /** \brief internal + */
        matrix4& operator+=(const double& s);
        /** \brief internal - */
        matrix4& operator-=(const matrix4& m);
        /** \brief internal - */
        matrix4& operator-=(const double& s);
        /** \brief internal * */
        matrix4& operator*=(const double& s);
        /** \brief internal / */
        matrix4& operator/=(const double& s);

        /** \brief unary negation */
        matrix4 operator-() const;

        /** \brief does componentwise mutliplication
        \return MC_v2d(x*p.x(),y*p.y(),z*p.z())
    */
        matrix4 product_compontentwise(const matrix4& m) const;
        /** \brief does componentwise mutliplication

        perform p=(p.x()*p2.x(),p.y()*p2.x())
    */
        matrix4& product_compontentwise_internal(const matrix4& m);

        /** \brief does the product a^t M b, where (a,b) are v4 vectors */
        double scalar_product(const v4& a,const v4& b) const;

        /** \brief transpose matrix */
        matrix4 transposed() const;


        // ********************************************* //
        // ********************************************* //
        //  Output
        // ********************************************* //
        // ********************************************* //

        /** \brief output the vector in ostream */
        friend std::ostream& operator<<(std::ostream& stream,const matrix4& m);





    private:

        // ********************************************* //
        // ********************************************* //
        //  Internal parameters
        // ********************************************* //
        // ********************************************* //

        /** \brief internal storage of the matrix */
        double m[16];


        // ********************************************* //
        // ********************************************* //
        //  Helper
        // ********************************************* //
        // ********************************************* //

        /** \brief assert that a size_t belongs to [[0,3]] */
        void assert_size(const size_t& k1,const size_t& k2) const throw(std::exception);

};
}

#endif
