/*
**    TP 4ETI CPE Lyon
**    Copyright (C) 2012 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef _V4_HPP_
#define _V4_HPP_

#include <iostream>

namespace cpe
{

    class v3;

    /** \brief 4D point */
    class v4
    {
    public:

        // ********************************************* //
        // ********************************************* //
        //  CONSTRUCTORS
        // ********************************************* //
        // ********************************************* //

        /** \brief empty constructor */
        v4();
        /** \brief direct constructor */
        v4(const double& x,const double& y,const double& z,const double& w);
        /** \brief constructor from v3 & double concatenation */
        v4(const v3& xyz,const double& w);

        // ********************************************* //
        // ********************************************* //
        //  Get/set
        // ********************************************* //
        // ********************************************* //

        /** \brief get x coordinate */
        const double& x() const;
        /** \brief get x coordinate */
        double& x();
        /** \brief get y coordinate */
        const double& y() const;
        /** \brief get y coordinate */
        double& y();
        /** \brief get z coordinate */
        const double& z() const;
        /** \brief get z coordinate */
        double& z();
        /** \brief get w coordinate */
        const double& w() const;
        /** \brief get z coordinate */
        double& w();

        /** \brief Access to the k_th entry (k in [0,3]) */
        const double& operator[](const size_t& k) const;
        /** \brief Access to the k_th entry (k in [0,3]) */
        double& operator[](const size_t& k);
        /** \brief Access to the k_th entry (k in [0,3]) */
        const double& operator()(const size_t& k) const;
        /** \brief Access to the k_th entry (k in [0,3]) */
        double& operator()(const size_t& k);


        /** \brief set every entry to 0 */
        void set_zero();


        // ********************************************* //
        // ********************************************* //
        //  Math operation
        // ********************************************* //
        // ********************************************* //

        /** \brief perform dot product between two v4 */
        double dot(const v4& p) const;

        /** \brief get the norm of the vector */
        double norm() const;
        /** \brief get the square norm of the vector */
        double norm2() const;






        // ********************************************* //
        // ********************************************* //
        //  Operator +-*/
        // ********************************************* //
        // ********************************************* //

        /** \brief + operator */
        friend v4 operator+(const v4& p1,const v4& p2);
        /** \brief + operator */
        friend v4 operator+(const double& s,const v4& p);
        /** \brief + operator */
        friend v4 operator+(const v4& p,const double& s);

        /** \brief - operator */
        friend v4 operator-(const v4& p1,const v4& p2);
        /** \brief - operator */
        friend v4 operator-(const double& s,const v4& p);
        /** \brief - operator */
        friend v4 operator-(const v4& p,const double& s);

        /** \brief multiply by a scalar operator */
        friend v4 operator*(const double& s,const v4& p);
        /** \brief multiply by a scalar operator */
        friend v4 operator*(const v4& p,const double& s);
        /** \brief divide by a scalar operator */
        friend v4 operator/(const v4& p,const double& s);

        /** \brief internal + */
        v4& operator+=(const v4& p);
        /** \brief internal + */
        v4& operator+=(const double& s);
        /** \brief internal - */
        v4& operator-=(const v4& p);
        /** \brief internal - */
        v4& operator-=(const double& s);
        /** \brief internal * */
        v4& operator*=(const double& s);
        /** \brief internal / */
        v4& operator/=(const double& s);

        /** \brief unary negation */
        v4 operator-() const;

        /** \brief does componentwise mutliplication
          \return MC_v2d(x*p.x(),y*p.y(),z*p.z())
        */
        v4 product_compontentwise(const v4& p) const;
        /** \brief does componentwise mutliplication

          perform p=(p.x()*p2.x(),p.y()*p2.x())
        */
        v4& product_compontentwise_internal(const v4& p);


        // ********************************************* //
        // ********************************************* //
        //  Output
        // ********************************************* //
        // ********************************************* //

        /** \brief export the value as string cout<<v4(2,3,6,7) => 2 3 6 7 */
        std::string to_string() const;

        /** \brief output the vector in ostream as (x,y,z,w) */
        friend std::ostream& operator<<(std::ostream& stream,const v4& p);



    private:

        // ********************************************* //
        // ********************************************* //
        //  Internal parameters
        // ********************************************* //
        // ********************************************* //

        /** \brief x coordinate */
        double internal_x;
        /** \brief y coordinate */
        double internal_y;
        /** \brief z coordinate */
        double internal_z;
        /** \brief w coordinate */
        double internal_w;


        // ********************************************* //
        // ********************************************* //
        //  Helper
        // ********************************************* //
        // ********************************************* //

        /** \brief assert that a size_t belongs to [[0,3]] */
        void assert_size(const size_t& k) const throw(std::exception);


    };

}

#endif
