/*
**    TP 4ETI CPE Lyon
**    Copyright (C) 2012 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <matrix4.hpp>
#include <v3.hpp>
#include <v4.hpp>
#include <cmath>
#include <string_helper.hpp>

namespace cpe
{


matrix4::matrix4()
{
    m[0]=1;m[4]=0;m[8 ]=0;m[12]=0;
    m[1]=0;m[5]=1;m[9 ]=0;m[13]=0;
    m[2]=0;m[6]=0;m[10]=1;m[14]=0;
    m[3]=0;m[7]=0;m[11]=0;m[15]=1;
}

matrix4::matrix4(const double& x00,const double& x01,const double& x02,const double& x03,
                 const double& x10,const double& x11,const double& x12,const double& x13,
                 const double& x20,const double& x21,const double& x22,const double& x23,
                 const double& x30,const double& x31,const double& x32,const double& x33)
{
    m[0]=x00;m[4]=x01;m[8 ]=x02;m[12]=x03;
    m[1]=x10;m[5]=x11;m[9 ]=x12;m[13]=x13;
    m[2]=x20;m[6]=x21;m[10]=x22;m[14]=x23;
    m[3]=x30;m[7]=x31;m[11]=x32;m[15]=x33;
}
matrix4 matrix4::identity()
{
    return matrix4(1,0,0,0 , 0,1,0,0 , 0,0,1,0 , 0,0,0,1);
}
matrix4 matrix4::zeros()
{
    return matrix4(0,0,0,0 , 0,0,0,0 , 0,0,0,0 , 0,0,0,0);
}


matrix4 matrix4::rotation(const v3& axis,const double& angle)
{
    v3 n=axis.normalized();

    double cos_t=std::cos(angle);
    double sin_t=std::sin(angle);

    return matrix4(cos_t+n[0]*n[0]*(1-cos_t),n[0]*n[1]*(1-cos_t)-n[2]*sin_t,n[1]*sin_t+n[0]*n[2]*(1-cos_t),0,
                   n[2]*sin_t+n[0]*n[1]*(1-cos_t),cos_t+n[1]*n[1]*(1.0-cos_t),-n[0]*sin_t+n[1]*n[2]*(1-cos_t),0,
                   -n[1]*sin_t+n[0]*n[2]*(1-cos_t),n[0]*sin_t+n[1]*n[2]*(1-cos_t),cos_t+n[2]*n[2]*(1-cos_t),0,
                   0,0,0,1);

}

matrix4 matrix4::scale(const double& s)
{
    return matrix4(s,0,0,0 , 0,s,0,0 , 0,0,s,0 , 0,0,0,1);
}

matrix4 matrix4::scale(const double& s_x,const double& s_y,const double& s_z,const double& s_w)
{
    return matrix4(s_x,0,0,0 , 0,s_y,0,0 , 0,0,s_z,0 , 0,0,0,s_w);
}

const double& matrix4::operator()(const size_t& k1,const size_t& k2) const
{
    assert_size(k1,k2);
    return m[4*k2+k1];
}
double& matrix4::operator()(const size_t& k1,const size_t& k2)
{
    assert_size(k1,k2);
    return m[4*k2+k1];
}

matrix4 operator+(const matrix4& m1,const matrix4& m2)
{
    const double* pm1=m1.m;
    const double* pm2=m2.m;

    return matrix4(pm1[0]+pm2[0],pm1[4]+pm2[4],pm1[8 ]+pm2[8 ],pm1[12]+pm2[12],
                   pm1[1]+pm2[1],pm1[5]+pm2[5],pm1[9 ]+pm2[9 ],pm1[13]+pm2[13],
                   pm1[2]+pm2[2],pm1[6]+pm2[6],pm1[10]+pm2[10],pm1[14]+pm2[14],
                   pm1[3]+pm2[3],pm1[7]+pm2[7],pm1[11]+pm2[11],pm1[15]+pm2[15]);
}

matrix4 operator+(const double& s,const matrix4& _m)
{
    const double* pm=_m.m;
    return matrix4(pm[0]+s,pm[4]+s,pm[8 ]+s,pm[12]+s,
                   pm[1]+s,pm[5]+s,pm[9 ]+s,pm[13]+s,
                   pm[2]+s,pm[6]+s,pm[10]+s,pm[14]+s,
                   pm[3]+s,pm[7]+s,pm[11]+s,pm[15]+s);
}

matrix4 operator+(const matrix4& m,const double& s)
{
    return s+m;
}

matrix4 operator-(const matrix4& m1,const matrix4& m2)
{
    const double* pm1=m1.m;
    const double* pm2=m2.m;

    return matrix4(pm1[0]-pm2[0],pm1[4]-pm2[4],pm1[8 ]-pm2[8 ],pm1[12]-pm2[12],
                   pm1[1]-pm2[1],pm1[5]-pm2[5],pm1[9 ]-pm2[9 ],pm1[13]-pm2[13],
                   pm1[2]-pm2[2],pm1[6]-pm2[6],pm1[10]-pm2[10],pm1[14]-pm2[14],
                   pm1[3]-pm2[3],pm1[7]-pm2[7],pm1[11]-pm2[11],pm1[15]-pm2[15]);
}

matrix4 operator-(const double& s,const matrix4& _m)
{
    const double* pm=_m.m;
    return matrix4(s-pm[0],s-pm[4],s-pm[8 ],s-pm[12],
                   s-pm[1],s-pm[5],s-pm[9 ],s-pm[13],
                   s-pm[2],s-pm[6],s-pm[10],s-pm[14],
                   s-pm[3],s-pm[7],s-pm[11],s-pm[15]);
}

matrix4 operator-(const matrix4& _m,const double& s)
{
    const double* pm=_m.m;
    return matrix4(pm[0]-s,pm[4]-s,pm[8 ]-s,pm[12]-s,
                   pm[1]-s,pm[5]-s,pm[9 ]-s,pm[13]-s,
                   pm[2]-s,pm[6]-s,pm[10]-s,pm[14]-s,
                   pm[3]-s,pm[7]-s,pm[11]-s,pm[15]-s);
}

matrix4 operator*(const double& s,const matrix4& _m)
{
    const double* pm=_m.m;
    return matrix4(pm[0]*s,pm[4]*s,pm[8 ]*s,pm[12]*s,
                   pm[1]*s,pm[5]*s,pm[9 ]*s,pm[13]*s,
                   pm[2]*s,pm[6]*s,pm[10]*s,pm[14]*s,
                   pm[3]*s,pm[7]*s,pm[11]*s,pm[15]*s);
}

matrix4 operator*(const matrix4& m,const double& s)
{
    return s*m;
}

v4 operator*(const matrix4& _m,const v4& v)
{
    const double* pm=_m.m;
    double x0=v.x();
    double y0=v.y();
    double z0=v.z();
    double w0=v.w();

    return v4(pm[0]*x0+pm[4]*y0+pm[8 ]*z0+pm[12]*w0,
              pm[1]*x0+pm[5]*y0+pm[9 ]*z0+pm[13]*w0,
              pm[2]*x0+pm[6]*y0+pm[10]*z0+pm[14]*w0,
              pm[3]*x0+pm[7]*y0+pm[11]*z0+pm[15]*w0);

}

v4 operator*(const v4& v,const matrix4& _m)
{
    const double* pm=_m.m;
    double x0=v.x();
    double y0=v.y();
    double z0=v.z();
    double w0=v.w();

    return v4(pm[0 ]*x0+pm[1 ]*y0+pm[2 ]*z0+pm[3 ]*w0,
              pm[4 ]*x0+pm[5 ]*y0+pm[6 ]*z0+pm[7 ]*w0,
              pm[8 ]*x0+pm[9 ]*y0+pm[10]*z0+pm[11]*w0,
              pm[12]*x0+pm[13]*y0+pm[14]*z0+pm[15]*w0);

}

matrix4 operator*(const matrix4& m1,const matrix4& m2)
{
    const double* pm1=m1.m;
    const double* pm2=m2.m;

    matrix4 current=matrix4::zeros();
    double *p=current.pointer_unprotected();
    for(unsigned int k0=0;k0<4;++k0)
        for(unsigned int k1=0;k1<4;++k1)
            for(unsigned int ks=0;ks<4;++ks)
                p[k0+4*k1] += pm1[4*ks+k0]*pm2[4*k1+ks];
    return current;
}

matrix4 operator/(const matrix4& _m,const double& s)
{
    const double* pm=_m.m;
    return matrix4(pm[0]/s,pm[4]/s,pm[8 ]/s,pm[12]/s,
                   pm[1]/s,pm[5]/s,pm[9 ]/s,pm[13]/s,
                   pm[2]/s,pm[6]/s,pm[10]/s,pm[14]/s,
                   pm[3]/s,pm[7]/s,pm[11]/s,pm[15]/s);
}

matrix4& matrix4::operator+=(const matrix4& _m)
{
    const double* pm=_m.m;
    for(unsigned int k=0;k<16;++k)
        m[k] += pm[k];
    return *this;
}

matrix4& matrix4::operator+=(const double& s)
{
    for(unsigned int k=0;k<16;++k)
        m[k] += s;
    return *this;
}

matrix4& matrix4::operator-=(const matrix4& _m)
{
    const double* pm=_m.m;
    for(unsigned int k=0;k<16;++k)
        m[k] -= pm[k];
    return *this;
}

matrix4& matrix4::operator-=(const double& s)
{
    for(unsigned int k=0;k<16;++k)
        m[k] -= s;
    return *this;
}

matrix4& matrix4::operator*=(const double& s)
{
    for(unsigned int k=0;k<16;++k)
        m[k] *= s;
    return *this;
}

matrix4& matrix4::operator/=(const double& s)
{
    for(unsigned int k=0;k<16;++k)
        m[k] /= s;
    return *this;
}

matrix4 matrix4::operator-() const
{
    return matrix4(-m[0],-m[4],-m[8 ],-m[12],
                   -m[1],-m[5],-m[9 ],-m[13],
                   -m[2],-m[6],-m[10],-m[14],
                   -m[3],-m[7],-m[11],-m[15]);
}
matrix4 matrix4::product_compontentwise(const matrix4& _m) const
{
    const double* pm=_m.m;
    return matrix4(m[0]*pm[0],m[4]*pm[4],m[8 ]*pm[8 ],m[12]*pm[12],
                   m[1]*pm[1],m[5]*pm[5],m[9 ]*pm[9 ],m[13]*pm[13],
                   m[2]*pm[2],m[6]*pm[6],m[10]*pm[10],m[14]*pm[14],
                   m[3]*pm[3],m[7]*pm[7],m[11]*pm[11],m[15]*pm[15]);
}

matrix4& matrix4::product_compontentwise_internal(const matrix4& _m)
{
    const double* pm=_m.m;
    for(unsigned int k=0;k<16;++k)
        m[k]*=pm[k];
    return *this;
}

std::ostream& operator<<(std::ostream& stream,const matrix4& _m)
{
    const double* pm=_m.m;
    stream<<pm[0]<<","<<pm[4]<<","<<pm[8 ]<<","<<pm[12]<<"\n"<<
            pm[1]<<","<<pm[5]<<","<<pm[9 ]<<","<<pm[13]<<"\n"<<
            pm[2]<<","<<pm[6]<<","<<pm[10]<<","<<pm[14]<<"\n"<<
            pm[3]<<","<<pm[7]<<","<<pm[11]<<","<<pm[15]<<"\n";
    return stream;
}

void matrix4::assert_size(const size_t& k1,const size_t& k2) const throw(std::exception)
{
    if(k1>3 || k2>3)
        throw std::string("matrix4::assert_size("+cpe::string_converter::to_string(k1)+","+cpe::string_converter::to_string(k2)+") should be: k1<4 & k2<4");
}

const double* matrix4::pointer() const
{
    return m;
}
double* matrix4::pointer_unprotected()
{
    return m;
}

matrix4 matrix4::transposed() const
{
    return matrix4(m[0 ],m[1 ],m[2 ],m[3 ],
                   m[4 ],m[5 ],m[6 ],m[7 ],
                   m[8 ],m[9 ],m[10],m[11],
                   m[12],m[13],m[14],m[15]);
}

double matrix4::scalar_product(const v4& a,const v4& b) const
{
    return a.dot((*this)*b);
}

}
